package com.fundi.framework.common.helpers;

/*
 *  Common Services Library - Framework plug-in.
 *  (C) Copyright Fundi Software 2011, 2016.  Fundi Confidential.
 *  All rights reserved.  Use, duplication or disclosure restricted. 
 */

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.Marshaller;
import javax.xml.bind.Unmarshaller;

import com.fundi.framework.common.displaylist.DisplayList;
import com.fundi.framework.common.displaylist.ListLayout;
import com.fundi.framework.common.displaylist.ListLayoutV1;
import com.fundi.framework.common.instance.AppInstance;
import com.fundi.framework.common.nl1.Messages;

public class ListLayoutHelper implements IListLayoutHelper {
	
	//Eye catchers.
	public static final String EC_COPYRIGHT = "(C) Copyright 2011, 2016 Fundi Software.  Fundi Confidential."; //$NON-NLS-1$
	public static final String EC_APAR = "3.1.1.00_00014"; //$NON-NLS-1$
	public static final String EC_FMID = "H30SA10"; //$NON-NLS-1$
	public static final String EC_PID = "5655-CSL"; //$NON-NLS-1$

	private static int					version = 1;

	private Map<String, ListLayoutV1>	layoutMap;
	private AppInstance					instance;
	private boolean						upgrade = true;
	private String						listLayoutFileNamePrefix = Messages.getString("ListLayoutHelper_0"); //$NON-NLS-1$

	public ListLayoutHelper(AppInstance instance) {
		layoutMap = new HashMap<String, ListLayoutV1>() {
			public ListLayoutV1 put(String key, ListLayoutV1 layout) {
				if (layout == null) (new Exception(Messages.getString("ListLayoutHelper_1") + key)).printStackTrace(); //$NON-NLS-1$
				
				return super.put(key, layout);
			}
		};
		this.instance = instance;
	}

	@Override
	public ListLayoutV1 get(String context) {
		ListLayoutV1	result;
		
		if (context == null || context.isEmpty()) {
			result = new ListLayoutV1();	//these don't go into the map, this is to support lists which do not use layouts
		} else {
			result = layoutMap.get(context);
		}
		
		return result;
	}

	@Override
	public void set(DisplayList list) {
		ListLayoutV1 	layout;
		String 			context;
		
		try {
			context = list.getContext();
			if (context != null && !context.isEmpty()) {
				layout = load(context);
				layout.set(list);
			}
		} catch (Exception e) {
			instance.addConsoleMessage(e);
		}
	}
	
	@Override
	public ListLayoutV1 load(String context) {
		ListLayoutV1	result;
		
		result = get(context);
		if (result == null) {
			result = createLayout(context);
		}
		
		return result;
	}

	@Override
	public ListLayoutV1 load(DisplayList list) {
		ListLayoutV1	result;
		
		result = get(list.getContext());
		if (result == null) {
			result = createLayout(list);
		} else if (result.isEmpty()) {	//might have an excluded fields column definition and that is all, so pick up the data (and keys!) from the list.
			result.set(list);
		} else {
			result.ensureColumnLayoutsExist(list.getAllFields(), list);	//otherwise a new layout with query db followed by query db show(all) will not have the full column set in the config dialog.
		}
		
		return result;
	}

	@Override
	public void loadAll() {
		if (instance.savingListLayouts()) {
			upgradeAll();
			loadFromDisk(Messages.getString("ListLayoutHelper_2")); //$NON-NLS-1$
		}
	}
	
	@Override
	public void delete(String context) {
		deleteFromDisk(context);
		layoutMap.remove(context);
	}

	@Override
	public void save(String context) {
		saveToDisk(context);
	}
	
	@Override
	public void saveAll() {
		for (String context: layoutMap.keySet()) {
			save(context);
		}
	}

	@Override
	public ListLayoutV1 resetLayout(String context, DisplayList list) {
		ListLayoutV1	result;
		
		delete(context);
		result = createLayout(context);
		list.showAllColumns();
		list.setAscending(true);
		result.setDefaults(list);
		
		return result;
	}
	
	private ListLayoutV1 createLayout(String context) {
		ListLayoutV1	result;
		
		if (context == null || context.isEmpty()) (new Exception("Creating layout for null or empty context" + context)).printStackTrace(); //$NON-NLS-1$
		result = new ListLayoutV1(context);
		layoutMap.put(result.getContext(), result);
		
		return result;
	}

	private ListLayoutV1 createLayout(DisplayList list) {
		ListLayoutV1	result;
		
		result = createLayout(list.getContext());
		result.set(list);	//in this case, the new layout needs to take all its settings from the list, otherwise the list might get overridden - settings wiped.
		
		return result;
	}

	private synchronized void upgradeAll() {
		if (upgrade) {
			//check for the serialized ListLayout.txt. If found this needs to be read and then saved as a new set of XML files
			{
				String 					listLayoutsFilename = Messages.getString("ListLayoutHelper_3"); //$NON-NLS-1$
				ObjectInputStream 		in = null;
				ArrayList<ListLayout> 	listLayouts = new ArrayList<ListLayout>();
				File					legacyFile;
				ListLayoutV1			layoutV1;
				
				try {
					legacyFile = new File(instance.getWorkingFolder(), listLayoutsFilename);
					in = new ObjectInputStream(new FileInputStream(legacyFile));
					listLayouts = (ArrayList<ListLayout>) in.readObject();
					in.close();
					for (ListLayout layout: listLayouts) {
						layoutV1 = (ListLayoutV1) upgrade(layout);
						layoutMap.put(layoutV1.getContext(), layoutV1);	
					}
					saveAll();
					legacyFile.delete();
				} catch (FileNotFoundException fne) {
					//no action required
				} catch (Exception e) {
					instance.addConsoleMessage(e);
				}
			}
			upgrade = false;
		}
	}
	
	private Object upgrade(ListLayout legacyLayout) throws Exception {
		return upgrade((ListLayoutV1) new ListLayoutV1(legacyLayout));	//calling V1 upgrade now prevents this code requiring change (and consequent re-test) in future versions
	}
	
	private Object upgrade(ListLayoutV1 legacyLayout) {
		return legacyLayout;	//current version - nothing to do, future version - place code here
	}
	
	private void loadFromDisk(String context) {
		ListLayoutV1 	result;
		ArrayList<File> files;
		
		result = null;
		files = instance.getUserFiles(getFileName(context));
		for (File listLayoutFile: files) {
			try {
				result = loadFromStream(context, new FileInputStream(listLayoutFile));
				layoutMap.put(result.getContext(), result);
			} catch (Exception e) {
				instance.addConsoleMessage(e);
			}
		}
	}
	
	public ListLayoutV1 loadFromSnapshot(String context, byte[] xml) {
		ListLayoutV1 	result;
		
		result = null;
		try {
			result = loadFromStream(context, new ByteArrayInputStream(xml));
			layoutMap.put(result.getContext(), result);
		} catch (Exception e) {
			instance.addConsoleMessage(e);
		}
		
		return result;
	}
	
	private ListLayoutV1 loadFromStream(String context, InputStream in) {
		ListLayoutV1 	result;
		JAXBContext 	xmlContext;
		Unmarshaller 	u;
		
		result = null;
		try {
			xmlContext = JAXBContext.newInstance(ListLayoutV1.class);
			u = xmlContext.createUnmarshaller();
			result = (ListLayoutV1) u.unmarshal(in);
		} catch (Exception e) {
			instance.addConsoleMessage(e);
		}
		
		return result;
	}
	
	public void saveToDisk(String context) {
		File 				listLayoutFile;
		FileOutputStream 	out;
		
		if (context != null && !context.isEmpty()) {
			out = null;
			listLayoutFile = new File(instance.getWorkingFolder(), getFileName(context));
			try {
				out = new FileOutputStream(listLayoutFile.getAbsolutePath());
				saveToStream(context, out);
			} catch (Exception e) {
				instance.addConsoleMessage(e);
			} finally {
				if (out != null) {
					try {
						out.close();
					} catch (Exception e) {
						//nothing
					}
				}
			}
		}
	}
	
	public byte[] saveToSnapshot(String context) {
		ByteArrayOutputStream	out;	
		
		out = new ByteArrayOutputStream();
		try {
			saveToStream(context, out);
			//no point to calling close on a ByteArrayOutputStream
		} catch (Exception e) {
			instance.addConsoleMessage(e);
		}
		
		return out.toByteArray();
	}
	
	private void saveToStream(String context, OutputStream out) {
		JAXBContext 		xmlContext;
		Marshaller 			m;
		
		try {
			xmlContext = JAXBContext.newInstance(ListLayoutV1.class);
			m = xmlContext.createMarshaller();
			m.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, true);
			m.marshal(layoutMap.get(context), out);
		} catch (Exception e) {
			instance.addConsoleMessage(e);
		}
	}
	
	private void deleteFromDisk(String context) {
		ArrayList<File> files;
		
		if (context != null && !context.isEmpty()) {	//flexibility to allow delete all by allowing "" if desired, avoided for now
			files = instance.getUserFiles(getFileName(context));
			for (File listLayoutFile: files) {
				try {
					if (listLayoutFile.exists()) {
						listLayoutFile.delete();
					}
					layoutMap.remove(context);
				} catch (Exception e) {
					instance.addConsoleMessage(e);
				}
			}
		}
	}

	private String getFileName(String context) {
		return listLayoutFileNamePrefix + context + (context != null && !context.isEmpty() ? Messages.getString("ListLayoutHelper_4") : Messages.getString("ListLayoutHelper_5")); //$NON-NLS-1$ //$NON-NLS-2$
	}
}
