/*
 *  Common Services Library - Framework plug-in.
 *  (C) Copyright Fundi Software 2011, 2016.  Fundi Confidential.
 *  All rights reserved.  Use, duplication or disclosure restricted. 
 */

package com.fundi.framework.common.util;

import java.io.PrintStream;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

public class Tracer {
	
	//Eye catchers.
	public static final String EC_COPYRIGHT = "(C) Copyright 2011, 2016 Fundi Software.  Fundi Confidential."; //$NON-NLS-1$
	public static final String EC_APAR = "3.1.1.00_00014"; //$NON-NLS-1$
	public static final String EC_FMID = "H30SA10"; //$NON-NLS-1$
	public static final String EC_PID = "5655-CSL"; //$NON-NLS-1$

	Map<String, Stats> 	statsMap;
	PrintStream 		out;
	
	private static Tracer tracer = new Tracer();
	
	public static Tracer getInstance() {
		return tracer;
	}
	
	public static void remove() {
		Tracer tracer = getInstance();
		
		for (Stats stats: tracer.statsMap.values()) {
			stats.reset();
		}
		tracer.statsMap.clear();
		tracer.out.println("all stats removed"); //$NON-NLS-1$
	}

	public static void remove(String tag) {
		Tracer tracer = getInstance();
		
		Stats stats = tracer.getStats(tag);
		stats.reset();
		tracer.statsMap.remove(tag);
		tracer.out.println(tag + " removed"); //$NON-NLS-1$
	}
	
	public static void reset() {
		Tracer tracer = getInstance();
		
		for (Stats stats: tracer.statsMap.values()) {
			stats.reset();
		}
	}
	
	public static void reset(String tag) {
		Tracer tracer = getInstance();
		
		tracer.getStats(tag).reset();
	}
	
	public static void config(String tag, boolean trace, boolean accumulate, boolean time) {
		Tracer tracer = getInstance();
		
		tracer.getStats(tag).ensureStrategies(trace, accumulate, time);
	}
	
	public static void start(String tag) {
		Tracer tracer = getInstance();
		Stats stats = tracer.getStats(tag);
		
		stats.ensureStrategies(false, false, true);
		stats.enter();
	}

	public static void accumulate(String tag) {
		Tracer tracer = getInstance();
		Stats stats = tracer.getStats(tag);
		
		stats.ensureStrategies(false, true, false);
		stats.enter();
	}
	
	public static void trace(String tag) {
		Tracer tracer = getInstance();
		Stats stats = tracer.getStats(tag);
		
		stats.ensureStrategies(true, false, false);
		stats.enter();
	}
	
	public static void enter(String tag) {
		Tracer tracer = getInstance();
		
		tracer.getStats(tag).enter();
	}
	
	public static void exit(String tag) {
		Tracer tracer = getInstance();
		
		tracer.getStats(tag).exit();
	}
	

	
	public Tracer() {
		statsMap = new HashMap<String, Stats>();
		out = System.out;
	}
	
	private Stats getStats(String tag) {
		Stats result;
		
		result = statsMap.get(tag);
		if (result == null) {
			result = new Stats(this, tag);
			statsMap.put(tag, result);
		}
		
		return result;
	}
	

	
	//CLASS DEFINITION/////////////////////////////////////////////////////////////////////
	public class Stats {
		private String 			tag;
		private List<Strategy>	strategies;
		private Tracer			tracer;
		//
		public Stats(Tracer tracer, String tag) {
			this(tracer, tag, false, false, false);
		}
		public Stats(Tracer tracer, String tag, boolean trace, boolean accumulate, boolean time) {
			this.tracer = tracer;
			this.tag = tag;
			this.strategies = new LinkedList<Strategy>();
			ensureStrategies(trace, accumulate, time);
//			out.println(tag + " created");
		}
		public void ensureStrategies(boolean trace, boolean accumulate, boolean time) {
			boolean addTrace = trace;
			boolean addAccumulate = accumulate;
			boolean addTime = time;
			for (Strategy strategy: strategies) {
				if (strategy.getClass() == Trace.class){
					addTrace = false;
				}
				if (strategy.getClass() == Accumulate.class){
					addAccumulate = false;
				}
				if (strategy.getClass() == Time.class){
					addTime = false;
				}
			}
			if (addTrace) {
				strategies.add(new Trace(this, tracer.out));
//				out.println(tag + " added Trace");
			}
			if (addAccumulate) {
				strategies.add(new Accumulate(this, tracer.out));
//				out.println(tag + " added Accumulate");
			}
			if (addTime) {
				strategies.add(new Time(this, tracer.out));
//				out.println(tag + " added Time");
			}
		}
		public void addStrategy(Strategy strategy) {
			if (!strategies.contains(strategy)) {
				strategies.add(strategy);
			}
		}
		public void removeStrategy(Strategy strategy) {
			strategies.remove(strategy);
		}
		public void enter() {
			for (Strategy strategy: strategies) {
				strategy.enter();
			}
		}
		public void exit() {
			for (Strategy strategy: strategies) {
				strategy.exit();
			}
		}
		public void reset() {
			for (Strategy strategy: strategies) {
				strategy.reset();
			}
		}
	}



	//CLASS DECLARATION/////////////////////////////////////////////////////////////////////
	public interface Strategy {
		public void enter();
		public void exit();
		public void reset();
	}
	
	
	
	//CLASS DECLARATION/////////////////////////////////////////////////////////////////////
	public class Trace implements Strategy {
		Stats		parent;
		PrintStream	out;
		//
		public Trace(Stats parent, PrintStream out) {
			this.parent = parent;
			this.out = out;
		}
		@Override
		public void enter() {
			out.println(parent.tag + " enter>"); //$NON-NLS-1$
		}
		@Override
		public void exit() {
			out.println(parent.tag + " <exit"); //$NON-NLS-1$
		}
		@Override
		public void reset() {}	//nothing to do
		@Override
		public boolean equals(Object o) {
			return this.getClass() == o.getClass();
		}
	}
	
	
	
	//CLASS DECLARATION/////////////////////////////////////////////////////////////////////
	public class Accumulate implements Strategy {
		Stats		parent;
		PrintStream	out;
		long		sum;
		long		lastEnter;
		long		lastExit;
		//
		public Accumulate(Stats parent, PrintStream out) {
			this.parent = parent;
			this.out = out;
			sum = 0;
			lastEnter = -1;
			lastExit = 0;
		}
		@Override
		public void enter() {
			if (lastEnter <= lastExit) { //then consider this to be a new timing interval
				lastEnter = System.currentTimeMillis();
			} //else, leave lastEnter alone as this looks like a recursion so this new 'timing' is within the existing 'timing'
		}
		@Override
		public void exit() {
			long exit = System.currentTimeMillis();
			
			if (lastEnter >= lastExit) {
				lastExit = exit;
				sum += lastExit - lastEnter;
			} else { //only increment by the elapsed time since lastExit
				sum += exit - lastExit;
				lastExit = exit;
			}
		}
		@Override
		public void reset() {
			out.println(parent.tag + " total time: " + sum + "ms"); //$NON-NLS-1$ //$NON-NLS-2$
			sum = 0;
			lastEnter = -1;
			lastExit = 0;
		}
		@Override
		public boolean equals(Object o) {
			return this.getClass() == o.getClass();
		}
	}
	
	
	
	//CLASS DECLARATION/////////////////////////////////////////////////////////////////////
	public class Time implements Strategy {
		Stats		parent;
		PrintStream	out;
		long		lastEnter;
		//
		public Time(Stats parent, PrintStream out) {
			this.parent = parent;
			this.out = out;
			lastEnter = -1;
		}
		@Override
		public void enter() {
			if (lastEnter != -1) {
				Throwable t = new Throwable();
				out.print(parent.tag + " ***Error: repeat enter call without corresponding exit not accumulated*** worth " + (System.currentTimeMillis() - lastEnter) + "ms at "); //$NON-NLS-1$ //$NON-NLS-2$
				t.printStackTrace(out);
			}
			lastEnter = System.currentTimeMillis();
		}
		@Override
		public void exit() {
			if (lastEnter == -1) {
				Throwable t = new Throwable();
				out.print(parent.tag + " ***Error: attempt to record exit with no corresponding enter*** at "); //$NON-NLS-1$
				t.printStackTrace(out);
			} else {
				reset();
			}
		}
		@Override
		public void reset() {
			out.println(parent.tag + " time: " + (System.currentTimeMillis() - lastEnter) + "ms"); //$NON-NLS-1$ //$NON-NLS-2$
			lastEnter = -1;
		}
		@Override
		public boolean equals(Object o) {
			return this.getClass() == o.getClass();
		}
	}
	
	
	
//	//CLASS DEFINITION/////////////////////////////////////////////////////////////////////
//	public abstract class AccBlock {
//		public String tag;
//		//
//		public AccBlock(String tag) {
//			this.tag = tag;
//			config(tag, false, true, false);
//			preRun();
//			run();
//			postRun();
//		}
//		private void preRun() {
//			Tracer.enter(tag);
//		}
//		/** 
//		 * 	wrap this run method around the code to be traced
//		 */
//		public abstract void run();
//		private void postRun() {
//			Tracer.exit(tag);
//		}
//	}
//
//
//
//	//CLASS DEFINITION/////////////////////////////////////////////////////////////////////
//	public abstract class TrcBlock {
//		public String tag;
//		//
//		public TrcBlock(String tag) {
//			this.tag = tag;
//			config(tag, true, false, false);
//			preRun();
//			run();
//			postRun();
//		}
//		private void preRun() {
//			Tracer.enter(tag);
//		}
//		/** 
//		 * 	wrap this run method around the code to be traced
//		 */
//		public abstract void run();
//		private void postRun() {
//			Tracer.exit(tag);
//		}
//	}
}
